'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _localFs = require('./local-fs');

var _localFs2 = _interopRequireDefault(_localFs);

var _async = require('async');

var _async2 = _interopRequireDefault(_async);

var _mkdirp = require('mkdirp');

var _mkdirp2 = _interopRequireDefault(_mkdirp);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Handle local database.
 */

// $FlowFixMe
class LocalDatabase {

  /**
   * Load an parse the local json database.
   * @param {*} path the database path
   */
  constructor(config, logger) {
    this.config = config;
    this.path = this._buildStoragePath(config);
    this.logger = logger;
    this.locked = false;
    this.data = this._fetchLocalPackages();
    this._sync();
  }

  getSecret() {
    return Promise.resolve(this.data.secret);
  }

  setSecret(secret) {
    return new Promise((resolve, reject) => {
      this.data.secret = secret;
      resolve(this._sync());
    });
  }

  /**
   * Add a new element.
   * @param {*} name
   * @return {Error|*}
   */
  add(name, cb) {
    if (this.data.list.indexOf(name) === -1) {
      this.data.list.push(name);
      cb(this._sync());
    } else {
      cb(null);
    }
  }

  search(onPackage, onEnd, validateName) {
    const storages = this._getCustomPackageLocalStorages();
    const base = _path2.default.dirname(this.config.self_path);
    const self = this;
    _async2.default.eachSeries(Object.keys(storages), function (storage, cb) {
      _fs2.default.readdir(_path2.default.resolve(base, storage), function (err, files) {
        if (err) {
          return cb(err);
        }

        _async2.default.eachSeries(files, function (file, cb) {
          if (file.match(/^@/)) {
            // scoped
            _fs2.default.readdir(_path2.default.resolve(base, storage, file), function (err, files) {
              if (err) {
                return cb(err);
              }

              _async2.default.eachSeries(files, (file2, cb) => {
                if (validateName(file2)) {
                  const packagePath = _path2.default.resolve(base, storage, file, file2);

                  _fs2.default.stat(packagePath, (err, stats) => {
                    if (_lodash2.default.isNil(err) === false) {
                      return cb(err);
                    }
                    const item = {
                      name: `${file}/${file2}`,
                      path: packagePath,
                      time: stats.mtime.getTime()
                    };
                    onPackage(item, cb);
                  });
                } else {
                  cb();
                }
              }, cb);
            });
          } else if (validateName(file)) {
            const packagePath = _path2.default.resolve(base, storage, file);
            _fs2.default.stat(packagePath, (err, stats) => {
              if (_lodash2.default.isNil(err) === false) {
                return cb(err);
              }
              onPackage({
                name: file,
                path: packagePath,
                time: self._getTime(stats.mtime.getTime(), stats.mtime)
              }, cb);
            });
          } else {
            cb();
          }
        }, cb);
      });
    }, onEnd);
  }

  _getTime(time, mtime) {
    return time ? time : mtime;
  }

  _getCustomPackageLocalStorages() {
    const storages = {};

    // add custom storage if exist
    if (this.config.storage) {
      storages[this.config.storage] = true;
    }

    const packages = this.config.packages;


    if (packages) {
      const listPackagesConf = Object.keys(packages || {});

      listPackagesConf.map(pkg => {
        if (packages[pkg].storage) {
          storages[packages[pkg].storage] = true;
        }
      });
    }

    return storages;
  }

  /**
   * Remove an element from the database.
   * @param {*} name
   * @return {Error|*}
   */
  remove(name, cb) {
    this.get((err, data) => {
      if (err) {
        cb(new Error('error on get'));
      }

      const pkgName = data.indexOf(name);
      if (pkgName !== -1) {
        this.data.list.splice(pkgName, 1);
      }

      cb(this._sync());
    });
  }

  /**
   * Return all database elements.
   * @return {Array}
   */
  get(cb) {
    cb(null, this.data.list);
  }

  /**
   * Syncronize {create} database whether does not exist.
   * @return {Error|*}
   */
  _sync() {
    if (this.locked) {
      this.logger.error('Database is locked, please check error message printed during startup to prevent data loss.');
      return new Error('Verdaccio database is locked, please contact your administrator to checkout logs during verdaccio startup.');
    }
    // Uses sync to prevent ugly race condition
    try {
      _mkdirp2.default.sync(_path2.default.dirname(this.path));
    } catch (err) {
      // perhaps a logger instance?
      return null;
    }

    try {
      _fs2.default.writeFileSync(this.path, JSON.stringify(this.data));
      return null;
    } catch (err) {
      return err;
    }
  }

  getPackageStorage(packageInfo) {
    // $FlowFixMe
    const packagePath = this._getLocalStoragePath(this.config.getMatchedPackagesSpec(packageInfo).storage);

    if (_lodash2.default.isString(packagePath) === false) {
      this.logger.debug({ name: packageInfo }, 'this package has no storage defined: @{name}');
      return;
    }

    const packageStoragePath = _path2.default.join(_path2.default.resolve(_path2.default.dirname(this.config.self_path || ''), packagePath), packageInfo);

    return new _localFs2.default(packageStoragePath, this.logger);
  }

  /**
   * Verify the right local storage location.
   * @param {String} path
   * @return {String}
   * @private
   */
  _getLocalStoragePath(path) {
    if (_lodash2.default.isNil(path) === false) {
      return path;
    }

    return this.config.storage;
  }

  /**
   * Build the local database path.
   * @param {Object} config
   * @return {string|String|*}
   * @private
   */
  _buildStoragePath(config) {
    return _path2.default.join(_path2.default.resolve(_path2.default.dirname(config.self_path || ''), config.storage, '.sinopia-db.json'));
  }

  /**
   * Fetch local packages.
   * @private
   * @return {Object}
   */
  _fetchLocalPackages() {
    const database = [];
    const emptyDatabase = { list: database, secret: '' };

    try {
      const dbFile = _fs2.default.readFileSync(this.path, 'utf8');

      if (_lodash2.default.isNil(dbFile)) {
        // readFileSync is platform specific, FreeBSD might return null
        return emptyDatabase;
      }

      const db = this._parseDatabase(dbFile);

      if (!db) {
        return emptyDatabase;
      }

      return db;
    } catch (err) {
      // readFileSync is platform specific, macOS, Linux and Windows thrown an error
      // Only recreate if file not found to prevent data loss
      if (err.code !== 'ENOENT') {
        this.locked = true;
        this.logger.error('Failed to read package database file, please check the error printed below:\n', `File Path: ${this.path}\n\n ${err.message}`);
      }
      return emptyDatabase;
    }
  }

  /**
   * Parse the local database.
   * @param {Object} dbFile
   * @private
   * @return {Object}
   */
  _parseDatabase(dbFile) {
    try {
      return JSON.parse(dbFile);
    } catch (err) {
      this.logger.error(`Package database file corrupted (invalid JSON), please check the error printed below.\nFile Path: ${this.path}`, err);
      this.locked = true;
    }
  }
}

exports.default = LocalDatabase;